# This project uses Poetry for dependency management.
import pytest
from pathlib import Path
import polars as pl
import pandas as pd
import openpyxl
from ez_excel_mgt import ExcelTemplate


def test_open_and_save_as(create_test_excel):
    """Test opening an existing file and saving as."""
    file_path, _, _ = create_test_excel
    template = ExcelTemplate(file_path)

    assert template is not None

    path = Path(file_path)
    template.save(str(path.with_suffix(".new.xlsx")))

    assert path.with_suffix(".new.xlsx").exists()

    # Clean up
    path.with_suffix(".new.xlsx").unlink()


def test_add_sheet(create_test_excel):
    """Test adding a sheet."""
    file_path, _, _ = create_test_excel
    dest_sheet_name = "Sheet 2"
    template = ExcelTemplate(str(file_path))

    template.add_sheet(dest_sheet_name)
    template.save(file_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(file_path)
    sheet = workbook[dest_sheet_name]

    assert sheet is not None

def test_list_sheet_names(create_test_excel_with_3_sheets):
    """Test listing sheet names."""
    file_path, sheet_names, _ = create_test_excel_with_3_sheets

    template = ExcelTemplate(file_path)
    
    assert template.sheet_names() == sheet_names

def test_write_cell(create_test_excel):
    """Test writing a cell."""
    file_path, _, _ = create_test_excel
    dest_sheet_name = "Test Sheet"

    template = ExcelTemplate(file_path)
    template.add_sheet(dest_sheet_name)
    template.write_cell(dest_sheet_name, (2, 5), "Hello, World!")
    template.write_cell(dest_sheet_name, "F2", 12345)
    template.save(file_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["E2"].value == "Hello, World!"
    assert sheet["F2"].value == 12345


def test_set_cell(create_test_excel):
    """Test setting a cell."""
    file_path, _, _ = create_test_excel
    dest_sheet_name = "Test Sheet"

    template = ExcelTemplate(file_path)
    template.add_sheet(dest_sheet_name)
    template.goto_sheet(dest_sheet_name)
    template.goto_cell((5, 2))
    template.set_value("Hello, World!")
    template.goto_cell('B6')
    template.set_value(True)
    template.save(file_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["B5"].value == "Hello, World!"
    assert sheet["B6"].value == True



import pytest
import openpyxl


@pytest.fixture(scope="function")
def create_test_excel(tmp_path):
    """Fixture to create a temporary Excel file for testing."""
    # Create a temporary file path for the Excel file
    excel_path = tmp_path / "test.xlsx"
    
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Sheet1"
    
    # Prepopulate the sheet with some data
    sheet["A1"] = "First row"
    sheet["A2"] = "Second row"
    sheet["A3"] = "Name"
    sheet["B3"] = "Age"
    sheet["C3"] = "Gender"
    sheet["A4"] = "Irène"
    sheet["B4"] = 25
    sheet["A5"] = "Matthieu"
    sheet["B5"] = 26

    # Write to the temporary file
    workbook.save(excel_path)
    
    return str(excel_path), sheet.title, 3


@pytest.fixture(scope="function")
def create_test_excel_float(tmp_path):
    """Fixture to create a temporary Excel file for testing."""
    # Create a temporary file path for the Excel file
    excel_path = tmp_path / "test.xlsx"
    
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Sheet1"
    
    # Prepopulate the sheet with some data
    sheet["A1"] = "First row"
    sheet["A2"] = "Second row"
    sheet["A3"] = "Name"
    sheet["B3"] = "Age"
    sheet["C3"] = "Gender"
    sheet["D3"] = "Revenue"
    sheet["A4"] = "Irène"
    sheet["B4"] = 25
    sheet["C4"] = "F"
    sheet["D4"] = 1234.56
    sheet["A5"] = "Matthieu"
    sheet["B5"] = 26
    sheet["C5"] = "M"
    sheet["D5"] = 9876.54

    # Write to the temporary file
    workbook.save(excel_path)
    
    return str(excel_path), sheet.title, 3


@pytest.fixture(scope="function")
def create_test_excel_commented(tmp_path):
    """Fixture to create a temporary Excel file for testing with comments after the header row."""
    # Create a temporary file path for the Excel file
    excel_path = tmp_path / "test.xlsx"
    
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Sheet1"
    
    # Prepopulate the sheet with some data
    sheet["A1"] = "First row"
    sheet["A2"] = "Second row"
    sheet["A3"] = "Name"
    sheet["B3"] = "Age"
    sheet["C3"] = "Gender"
    sheet["A4"] = "Commented name"
    sheet["B4"] = "Commented age"
    sheet["C4"] = "Commented gender"

    # Write to the temporary file
    workbook.save(excel_path)
    
    return str(excel_path), sheet.title, 3


@pytest.fixture(scope="function")
def create_empty_test_excel(tmp_path):
    """Fixture to create a temporary EmptyExcel file for testing."""
    # Create a temporary file path for the Excel file
    excel_path = "dest.xlsx"
    
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Test Sheet"
    
    # Write to the temporary file
    workbook.save(excel_path)
    
    return str(excel_path), sheet.title, 0


@pytest.fixture(scope="function")
def create_test_excel_with_data_to_aggregate(tmp_path):
    """Fixture to create a temporary Excel file for testing."""
    # Create a temporary file path for the Excel file
    excel_path = tmp_path / "test.xlsx"
    
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Sheet1"
    
    # Prepopulate the sheet with some data
    sheet["A1"] = "First"
    sheet["B1"] = "Second"
    sheet["C1"] = "Third"
    sheet["D1"] = "Fourth"
    sheet["A2"] = 1
    sheet["B2"] = 2
    sheet["C2"] = 3
    sheet["D2"] = 4
    sheet["A3"] = 5
    sheet["B3"] = 6
    sheet["C3"] = 7
    sheet["D3"] = 8

    # Write to the temporary file
    workbook.save(excel_path)
    
    return str(excel_path), sheet.title, 1


@pytest.fixture(scope="function")
def create_test_excel_with_3_sheets(tmp_path):
    """Fixture to create a temporary Excel file for testing with 3 sheets."""
    # Create a temporary file path for the Excel file
    excel_path = tmp_path / "test.xlsx"
    
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "New Sheet 1"
    workbook.create_sheet("New Sheet 2")
    workbook.create_sheet("New Sheet 3")
    
    # Write to the temporary file
    workbook.save(excel_path)
    
    return str(excel_path), ["New Sheet 1", "New Sheet 2", "New Sheet 3"], 3





from pathlib import Path
import re
from loguru import logger

# Map file extensions to their corresponding code block languages
EXTENSION_MAP = {
    ".py": "python",
    ".rs": "rust",
    ".java": "java",
    ".js": "javascript",
    ".html": "html",
    ".css": "css",
    ".sh": "bash",
    # Add more extensions and languages as needed
}

def md_include(source_md_file_path, target_md_file_path):
    # Convert the input path to a Path object
    source_md_file_path = Path(source_md_file_path)
    target_md_file_path = Path(target_md_file_path)

    # Regular expression to find !md_include directives
    include_pattern = re.compile(r'\s*!md_include\s*["\']([^"\']+)["\']')

    # Read the markdown file
    with source_md_file_path.open('r', encoding='utf-8') as md_file:
        md_content = md_file.readlines()

    # Process each line, looking for !md_include directive
    new_md_content = []
    for line in md_content:
        # Find all !md_include directives in the line
        matches = include_pattern.findall(line)

        if matches:
            for included_file_path in matches:
                # Make the path relative to the markdown file
                full_included_path = source_md_file_path.parent / included_file_path
                file_extension = full_included_path.suffix.lower()  # Get the file extension

                # Determine the code block language based on file extension
                language = EXTENSION_MAP.get(file_extension, "")  # Default to empty if not found

                # Read the content of the file to be included
                try:
                    with full_included_path.open('r', encoding='utf-8') as included_file:
                        included_content = included_file.read()

                        # Replace the !md_include directive with the content of the included file
                        # Wrap it in a code block with the appropriate language
                        if language:
                            line = line.replace(
                                f'!md_include "{included_file_path}"', 
                                f"```{language}\n{included_content}\n```"
                            )
                        else:
                            # No language specified, default to plain code block
                            line = line.replace(
                                f'!md_include "{included_file_path}"', 
                                f"```\n{included_content}\n```"
                            )
                except FileNotFoundError:
                    logger.error(f"Error: The file '{full_included_path}' could not be found.")
                    line = line.replace(f'!md_include "{included_file_path}"', f"**Error: Could not include file '{full_included_path}'**")

        new_md_content.append(line)

    # Write the updated content to a new file or overwrite the original
    with target_md_file_path.open('w', encoding='utf-8') as new_md_file:
        new_md_file.writelines(new_md_content)

    logger.info(f"Processed markdown file saved as: {target_md_file_path}")

# Example usage
if __name__ == "__main__":
    # Replace 'example.md' with the path to your actual markdown file
    md_include('docs/readme.md', 'README.md')


from typing import List, Dict, Any
from pathlib import Path

import pandas as pd
import polars as pl
import openpyxl
import ez_excel_mgt as ezex

# Create a template.xlsx file with a sheet named "Sheet1" and a header row at row 3
def create_template_excel(tmp_path: Path, 
                          sheet_name: str, 
                          metadata: List[str], 
                          data: Dict[str, List[Any]]) -> int:
    # Create a new workbook and add a sheet
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = sheet_name
    
    # Add metadata rows
    for i, row in enumerate(metadata):
        sheet[f"A{i+1}"] = row

    # Header row in row 3 (of the excel file)
    for j, (name, values) in enumerate(data.items()):
        sheet[f"{chr(65 + j)}{len(metadata)+1}"] = name
        for i, v in enumerate(values):
            sheet[f"{chr(65 + j)}{len(metadata)+i+1+1}"] = v

    # Write to the temporary file
    workbook.save(excel_path)
    
    return len(metadata) + 1


if __name__ == "__main__":
    sheet_name = "Example"
    excel_path = Path("example.xlsx")
    header_row = create_template_excel(excel_path,
                                       sheet_name, 
                                       ["A comment the first row", "Another in the second row"], 
                                       {"Name": ["Alice", "Bob", "Charlie"], 
                                        "Age": [25, 30, 35], 
                                        "Gender": ["F", "M", "M"],
                                        "City": ["New York", "London", "Paris"]})

    # Create a Pandas DataFrame
    pandas_df = pd.DataFrame({
        "Name": ["Anatole", "Erica", "Jules"],
        "Age": [85, 15, 95]
    })

    # Call the function to append the Pandas DataFrame to the existing sheet
    ezex.fill_sheet_with(pandas_df, str(excel_path), sheet_name, header_row=header_row)

    # Create a Polars DataFrame
    polars_df = pl.DataFrame({
        "Name": ["Philippe", "Paul"],
        "Age": [45, None],
        "Gender": ["M", "M"]
    })

    # Call the function to append the PolarsDataFrame to the existing sheet
    ezex.fill_sheet_with(polars_df, str(excel_path), sheet_name, header_row=header_row)

    # Create a dictionary with the data to be written to the Excel file
    dict_df = {
        "Name": ["Michel", "Amelie"],
        "Age": [35, 45],
        "Gender": ["M", "F"],
        "City": ["Paris", "London"]
    }

    # Read the Excel file into a Polars DataFrame
    ezex.fill_sheet_with(dict_df, str(excel_path), sheet_name, header_row=header_row)

    # Create a list of dictionaries with the data to be written to the Excel file
    # Column names must be provided separately
    columns = ["Age", "Gender", "City"]
    # skip_null allows to skip None values, which combined with overwrite enables filling gaps or replacing specific values like a mask
    list_df = [
        [None, None, None, None, None, None, None, 55, None, None],
        [None, None, None, "M", "F", "M", None, None, None, None],
        [None, None, None, "Brussels", "Madrid", "Berlin", "Lisbon", "Montreal", None, None]
    ]


    ezex.fill_sheet_with(list_df, str(excel_path), sheet_name, header_row=header_row, columns=columns, skip_null=True, overwrite=True)

    df = pl.read_excel(source=excel_path, sheet_name=sheet_name,
                       engine='xlsx2csv', engine_options={"skip_empty_lines": True, "skip_hidden_rows": False},
                       read_options={"skip_rows": 2, "has_header": True, "infer_schema_length": 0})
    print(df)


# This project uses Poetry for dependency management.
from pathlib import Path
import openpyxl
from ez_excel_mgt import ExcelTemplate


def test_transform_sum_row(create_test_excel_with_data_to_aggregate, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel_with_data_to_aggregate
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel

    template = ExcelTemplate(dest_file_path)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.aggregate_range_from(source_file_path, source_sheet_name, ((2, 1), (3, 4)), action='sum', mode='row')
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == 10
    assert sheet["A2"].value == 26

def test_transform_sum_col(create_test_excel_with_data_to_aggregate, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel_with_data_to_aggregate
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel
    
    template = ExcelTemplate(dest_file_path)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.aggregate_range_from(source_file_path, source_sheet_name, ((2, 1), (3, 4)), action='sum', mode='col')
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == 6
    assert sheet["B1"].value == 8
    assert sheet["C1"].value == 10
    assert sheet["D1"].value == 12

def test_transform_avg_row(create_test_excel_with_data_to_aggregate, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel_with_data_to_aggregate
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel

    template = ExcelTemplate(dest_file_path)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.aggregate_range_from(source_file_path, source_sheet_name, ((2, 1), (3, 4)), action='avg', mode='row')
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == 2.5
    assert sheet["A2"].value == 6.5

def test_transform_avg_col(create_test_excel_with_data_to_aggregate, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel_with_data_to_aggregate
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel


    template = ExcelTemplate(dest_file_path)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.aggregate_range_from(source_file_path, source_sheet_name, ((2, 1), (3, 4)), action='avg', mode='col')
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == 3
    assert sheet["B1"].value == 4
    assert sheet["C1"].value == 5
    assert sheet["D1"].value == 6



# This project uses Poetry for dependency management.
from pathlib import Path
import openpyxl
from ez_excel_mgt import ExcelTemplate


def test_copy_range_between_files(create_test_excel, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel

    template = ExcelTemplate(dest_file_path)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.copy_range_from(source_file_path, source_sheet_name, ((1, 1), (5, 3)), None, None)
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == "First row"
    assert sheet["A2"].value == "Second row"
    assert sheet["B4"].value == 25
    assert sheet["B5"].value == 26

def test_transpose_range_between_files(create_test_excel, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel

    template = ExcelTemplate(dest_file_path)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.copy_range_from(source_file_path, source_sheet_name, ((1, 1), (5, 3)), True, None)
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == "First row"
    assert sheet["B1"].value == "Second row"
    assert sheet["D2"].value == 25
    assert sheet["E2"].value == 26


def test_copy_range_between_files_and_coerce(create_test_excel_float, create_empty_test_excel):
    """Test copy."""
    source_file_path, source_sheet_name, _header_row = create_test_excel_float
    dest_file_path, dest_sheet_name, _header_row = create_empty_test_excel

    dest_sheet_name = "New Sheet"

    template = ExcelTemplate(dest_file_path)
    template.add_sheet(dest_sheet_name)
    template.goto_sheet(dest_sheet_name, cell=(1, 1))
    template.copy_range_from(source_file_path, source_sheet_name, ((4, 4), (5, 4)), None, 'int')
    template.save(dest_file_path)

    assert Path(dest_file_path).exists()

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(dest_file_path)
    sheet = workbook[dest_sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A1"].value == 1234
    assert sheet["A2"].value == 9876




# This project uses Poetry for dependency management.
 
import polars as pl
import pandas as pd
import pytest
import openpyxl
from ez_excel_mgt import ExcelTemplate


def generate_test_data(data, data_type="polars"):
    """Utility to generate test data in different formats (Polars, Pandas, or dict of lists)."""
    
    if data_type == "polars":
        return pl.DataFrame(data)
    elif data_type == "pandas":
        return pd.DataFrame(data)
    elif data_type == "dict":
        return data
    elif data_type == "list":
        return list(data.values())
    else:
        raise ValueError(f"Unsupported data type: {data_type}")
    

@pytest.mark.parametrize("data_type", ["polars", "pandas", "dict", "list"])
def test_fill_sheet(create_test_excel, data_type):
    """Test inserting data at the end of the sheet."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df = generate_test_data({
        "Name": ["Alice", "Bob"],
        "Age": [25, 30],
        "Gender": ["F", "M"],
    }, data_type)

    kwargs = {}
    if data_type == "list":
        kwargs["columns"] = ["Name", "Age", "Gender"]

    template = ExcelTemplate(excel_path)
    template.goto_sheet(sheet_name)
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df, **kwargs)
    template.save(excel_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(excel_path)
    sheet = workbook[sheet_name]
    
    # Assert that data is inserted with a header row (named columns)
    assert sheet["A3"].value == "Name"
    assert sheet["B3"].value == "Age"
    assert sheet["C3"].value == "Gender"
    assert sheet["A6"].value == "Alice"
    assert sheet["B6"].value == 25
    assert sheet["C6"].value == "F"
    assert sheet["A7"].value == "Bob"
    assert sheet["B7"].value == 30
    assert sheet["C7"].value == "M"

def test_fill_sheet_with_list_of_lists_with_no_columns(create_test_excel):
    """Test inserting data at the end of the sheet."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df = generate_test_data({
        "Name": ["Alice", "Bob"],
        "Age": [25, 30],
        "Gender": ["F", "M"]
    }, 'list')

    kwargs = {}

    with pytest.raises(ValueError, match="Column names must be provided for List of Lists."):
        template = ExcelTemplate(excel_path)
        template.goto_sheet(sheet_name)
        template.set_header_location((header_row, 1), 'row')
        template.fill_with(df, **kwargs)


def test_fill_sheet_with_list_of_lists_with_not_enough_columns(create_test_excel):
    """Test inserting data at the end of the sheet."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df = generate_test_data({
        "Name": ["Alice", "Bob"],
        "Age": [25, 30],
        "Gender": ["F", "M"]
    }, 'list')

    kwargs = {}
    kwargs["columns"] = ["Age", "Gender"]

    with pytest.raises(ValueError, match="List of columns and list of lists have different lengths."):
        template = ExcelTemplate(excel_path)
        template.goto_sheet(sheet_name)
        template.set_header_location((header_row, 1), 'row')
        template.fill_with(df, **kwargs)


@pytest.mark.parametrize("data_type", ["dict", "list"])
def test_fill_sheet_with_list_or_dict_of_lists_of_unequal_length(create_test_excel, data_type):
    """Test behavior when a column in the DataFrame is not of the same length as the others."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a DataFrame with mismatching column names
    df = generate_test_data({
        "Name": ["Alice", "Bob"],
        "Age": [25, 30, 54],
        "Gender": ["F", "M"]
    }, data_type)

    kwargs = {}
    if data_type == "list":
        kwargs["columns"] = ["Name", "Age", "Gender"]

    # Expect an error or handle the mismatch gracefully (depending on the implementation)
    message = "At least one list in the (dictionary|list) of lists has a different length than the others."
    with pytest.raises(ValueError, match=message):
        template = ExcelTemplate(excel_path)
        template.goto_sheet(sheet_name)
        template.set_header_location((header_row, 1), 'row')
        template.fill_with(df, **kwargs)


@pytest.mark.parametrize("data_type", ["pandas", "polars", "dict", "list"])
def test_fill_sheet_with_strict_and_missing_column(create_test_excel, data_type):
    """Test behavior when a column in the DataFrame is not of the same length as the others."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a DataFrame with mismatching column names
    df = generate_test_data({
        "Age": [25, 30],
        "Gender": ["F", "M"]
    }, data_type)

    kwargs = { "strict": True }
    if data_type == "list":
        kwargs["columns"] = ["Age", "Gender"]

    # Expect an error or handle the mismatch gracefully (depending on the implementation)
    message = "Header 'Name' in Sheet1 in the ExcelTemplate is missing in the DataFrame."
    with pytest.raises(ValueError, match=message):
        template = ExcelTemplate(excel_path)
        template.goto_sheet(sheet_name)
        template.set_header_location((header_row, 1), 'row')
        template.fill_with(df, **kwargs)


@pytest.mark.parametrize("data_type", ["pandas", "polars", "dict", "list"])
def test_fill_sheet_with_strict_and_unfound_column(create_test_excel, data_type):
    """Test behavior when a column in the DataFrame is not of the same length as the others."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a DataFrame with mismatching column names
    df = generate_test_data({
        "Full Name": ["Alice", "Bob"],
        "Age": [25, 30],
        "Gender": ["F", "M"]
    }, data_type)

    kwargs = { "strict": True }
    if data_type == "list":
        kwargs["columns"] = ["Full Name", "Age", "Gender"]

    # Expect an error or handle the mismatch gracefully (depending on the implementation)
    message = "Header 'Name' in Sheet1 in the ExcelTemplate is missing in the DataFrame."
    with pytest.raises(ValueError, match=message):
        template = ExcelTemplate(excel_path)
        template.goto_sheet(sheet_name)
        template.set_header_location((header_row, 1), 'row')
        template.fill_with(df, **kwargs)


@pytest.mark.parametrize("data_type", ["polars", "pandas", "dict", "list"])
def test_fill_sheet_with_overwrite_longer_than_template(create_test_excel, data_type):
    """Test overwriting data in the sheet."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df = generate_test_data({
        "Name": ["Alice", "Bob"],
        "Age": [25, 30],
        "Gender": ["F", "M"]
    }, data_type)

    kwargs = { "overwrite": True }
    if data_type == "list":
        kwargs["columns"] = ["Name", "Age", "Gender"]
        
    template = ExcelTemplate(excel_path)
    template.goto_sheet(sheet_name)
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df, **kwargs)
    template.save(excel_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(excel_path)
    sheet = workbook[sheet_name]
    
    # Assert that data is inserted with a header row (named columns)
    assert sheet["A3"].value == "Name"
    assert sheet["B3"].value == "Age"
    assert sheet["C3"].value == "Gender"
    assert sheet["A4"].value == "Alice"
    assert sheet["B4"].value == 25
    assert sheet["C4"].value == "F"
    assert sheet["A5"].value == "Bob"
    assert sheet["B5"].value == 30
    assert sheet["C5"].value == "M"


@pytest.mark.parametrize("data_type", ["polars", "pandas", "dict", "list"])
def test_fill_sheet_with_overwrite_shorter_than_template(create_test_excel, data_type):
    """Test overwriting data in the sheet."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df = generate_test_data({
        "Name": ["Alice"],
        "Age": [25],
        "Gender": ["F"]
    }, data_type)

    kwargs = { "overwrite": True }
    if data_type == "list":
        kwargs["columns"] = ["Name", "Age", "Gender"]
        
    template = ExcelTemplate(excel_path)
    template.goto_sheet(sheet_name)
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df, **kwargs)
    template.save(excel_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(excel_path)
    sheet = workbook[sheet_name]
    
    # Assert that data is inserted with a header row (named columns)
    assert sheet["A3"].value == "Name"
    assert sheet["B3"].value == "Age"
    assert sheet["C3"].value == "Gender"
    assert sheet["A4"].value == "Alice"
    assert sheet["B4"].value == 25
    assert sheet["C4"].value == "F"
    assert sheet["A5"].value == None
    assert sheet["B5"].value == None
    assert sheet["C5"].value == None


@pytest.mark.parametrize("data_type", ["polars", "pandas", "dict", "list"])
def test_fill_sheet_with_overwrite_and_skip_null(create_test_excel, data_type):
    """Test inserting a Polars DataFrame with named columns (`named=True`) and empty values."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df = generate_test_data({
        "Name": ["Alice", "Bob", "Tom"],
        "Age": [25, None, 30],
        "Gender": ["F", "M", None]
    }, data_type)
    
    kwargs = { "overwrite": True, "skip_null": True }
    if data_type == "list":
        kwargs["columns"] = ["Name", "Age", "Gender"]
        
    template = ExcelTemplate(excel_path)
    template.goto_sheet(sheet_name)
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df, **kwargs)
    template.save(excel_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(excel_path)
    sheet = workbook[sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A3"].value == "Name"
    assert sheet["B3"].value == "Age"
    assert sheet["C3"].value == "Gender"
    assert sheet["A4"].value == "Alice"
    assert sheet["B4"].value == 25
    assert sheet["C4"].value == "F"
    assert sheet["A5"].value == "Bob"
    assert sheet["B5"].value == 26
    assert sheet["C5"].value == "M"
    assert sheet["A6"].value == "Tom"
    assert sheet["B6"].value == 30
    assert sheet["C6"].value == None


def test_fill_sheet_with_empty_column_in_list(create_test_excel):
    """Test behavior when an empty column is in the DataFrame."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a DataFrame with mismatching column names
    df = generate_test_data({
        "Name": ["Alice", "Bob"],
        "Age": [25, 30],
        "Gender": []
    }, 'list')

    kwargs = {}
    kwargs["columns"] = ["Name", "Age", "Gender"]

    template = ExcelTemplate(excel_path)
    template.goto_sheet(sheet_name)
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df, **kwargs)
    template.save(excel_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(excel_path)
    sheet = workbook[sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A3"].value == "Name"
    assert sheet["B3"].value == "Age"
    assert sheet["C3"].value == "Gender"
    assert sheet["A6"].value == "Alice"
    assert sheet["B6"].value == 25
    assert sheet["C6"].value == None
    assert sheet["A7"].value == "Bob"
    assert sheet["B7"].value == 30
    assert sheet["C7"].value == None


@pytest.mark.parametrize("data_type", ["polars", "pandas", "dict", "list"])
def test_fill_sheet_with_multiple_overwrite(create_test_excel, data_type):
    """Test inserting data for specific columns in multiple calls."""
    excel_path, sheet_name, header_row = create_test_excel
    
    # Create a sample Polars DataFrame
    df1 = generate_test_data({
        "Name": ["Alice", "Bob", "Tom"],
        "Age": [25, 30, 35],
    }, data_type)

    kwargs1 = { "overwrite": True }
    if data_type == "list":
        kwargs1["columns"] = ["Name", "Age"]
        
    template = ExcelTemplate(excel_path)
    template.goto_sheet(sheet_name)
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df1, **kwargs1)

    df2 = generate_test_data({
        "Gender": ["F", "M", "M"]
    }, data_type)

    kwargs2 = { "overwrite": True }
    if data_type == "list":
        kwargs2["columns"] = ["Gender"]
        
    template.set_header_location((header_row, 1), 'row')
    template.fill_with(df2, **kwargs2)
    template.save(excel_path)

    # Load the modified Excel file and verify the contents
    workbook = openpyxl.load_workbook(excel_path)
    sheet = workbook[sheet_name]

    # Assert that data is inserted with a header row (named columns)
    assert sheet["A3"].value == "Name"
    assert sheet["B3"].value == "Age"
    assert sheet["C3"].value == "Gender"
    assert sheet["A4"].value == "Alice"
    assert sheet["B4"].value == 25
    assert sheet["C4"].value == "F"
    assert sheet["A5"].value == "Bob"
    assert sheet["B5"].value == 30
    assert sheet["C5"].value == "M"
    assert sheet["A6"].value == "Tom"
    assert sheet["B6"].value == 35
    assert sheet["C6"].value == "M"



import random
import polars as pl
import openpyxl
import ez_excel_mgt as ezex

# Function to convert a number to Excel column name
def excel_column(row: int, col: int) -> str:
    column_name = ""
    while row > 0:
        row, remainder = divmod(row - 1, 26)
        column_name = chr(65 + remainder) + column_name
    return f"{column_name}{col}"

if __name__ == "__main__":
    # Create a polars dataframe with 50 rows and 4 columns name Col 1, Col 2, Col 3, Col 4
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Example"
    for i in range(4):
        sheet[excel_column(i+1, 1)] = f"Col {i+1}"
    workbook.save("example_copy.xlsx")
    
    df = pl.DataFrame({
        "Col 1": [random.random() for _ in range(50)], 
        "Col 2": [random.random() for _ in range(50)], 
        "Col 3": [random.random() for _ in range(50)],
        "Col 4": [random.random() for _ in range(50)]
    })
    ezex.fill_sheet_with(df, "example_copy.xlsx", "Example")
    df = pl.read_excel(source="example_copy.xlsx", sheet_name="Example",
                       engine='xlsx2csv', engine_options={"skip_empty_lines": True, "skip_hidden_rows": False},
                       read_options={"has_header": True, "infer_schema_length": 0})
    print(df)

    # Open the destination excel file using openpyxl and create a new sheet named "Result", containing 50 headers in the first row named Row 1, Row 2, etc.
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "Result"

    # Update the loop to use the new function for naming columns
    for i in range(50):
        sheet[excel_column(i+1, 1)] = f"Row {i+1}"
    workbook.save("result_copy.xlsx")
    
    # Call the function to copy the range of cells from the source file to the destination file (row, col; starting at 1)
    ezex.copy_range_between_files("example_copy.xlsx", "Example", ((2, 1), (51, 4)), 
                                  "result_copy.xlsx", "Result", (2, 1), True)

    df = pl.read_excel(source="result_copy.xlsx", sheet_name="Result",
                       engine='xlsx2csv', engine_options={"skip_empty_lines": True, "skip_hidden_rows": False},
                       read_options={"has_header": True, "infer_schema_length": 0})
    print(df)